/*
* File: scanner.c
* ----------------
* This file implements the scanner.h interface.
*/

#include "stdafx.h"
#include "scanner.h"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

/*
* Type: ScannerImplementation
* ---------------------------
* This structure is the concrete representation of the type
* *Scanner, which is exported by this interface. The details
* of the representations are invisible to the client, but
* consist of the following fields:
*
* *file        -- Pointer to the file from which the tokens are
*                 read.
* *delimiters  -- String consisting of a set of delimiter
*                 characters supplied by the client.
* discardDelimiters -- Boolean value. If the value is true,
*                 the delimiter  will be discarded. Otherwise,
*                 the delimiter will be counted as a valid
*                 token.
* openedFile   -- Boolean value. This is set to true if
*                 NewScannerFromFilename is called, and false if
*                 NewScannerFromFile is called.
*/

struct ScannerImplementation
{
    FILE *file;
    char *delimiters;
    bool openedFile;
    bool discardDelimiters;
};

/* Private function prototypes */

static Scanner NewScanner(FILE *infile, const char *delimiters,
                          bool discardDelimiters, bool openedFile);
static bool ReadNonDelimiters(Scanner s, char buffer[], int bufLen);
static bool PutCharBack(Scanner s, char buffer[], int next, int i);
static int Skip(Scanner s, const char *set, bool over);
static bool IsInSet(char ch, const char *set);
static void* GetBlock(int nbytes);

/* Functions */

/*
* Function: NewScannerFromFilename
* ----------------------------------
* This function takes the filename from the client and asserts that it
* is not NULL. It then open the file, and passes a pointer to the file
* to NewScanner. Since this function opens the file, the 4th argument
* passed to NewScanner, openedFile, is set to true.
*/

Scanner NewScannerFromFilename(const char *filename, const char *delimiters,
                               bool discardDelimiters)
{
    FILE *infile;

    if (filename == NULL) return NULL;
    infile = fopen(filename, "r");
    return NewScanner(infile, delimiters, discardDelimiters, true);
}

/*
* Function: NewScannerFromFile
* -----------------------------
* This function passes all arguments to NewScanner. The 4th argument
* is set to false, because NewScannerFromFile does not open the file
* (*fp), unlike NewScannerFromFilename.
*/

Scanner NewScannerFromFile(FILE *fp, const char *delimiters,
                           bool discardDelimiters)
{
    return NewScanner(fp, delimiters, discardDelimiters, false);
}

/*
* Function: NewScanner
* ----------------------
* This function creates a new scanner, and returns a pointer to it.
* It first asserts that there are delimiter characters. If the file
* passed in is NULL, the scanner is returned a NULL pointer.
* If delimiters nor infile is not NULL, the function allocates a memory
* in the heap, of size ScanImplementation(*scan). It copies all arguments
* passed into this function to the corresponding variables in the
* structure.
*/

static Scanner NewScanner(FILE *infile, const char *delimiters,
                          bool discardDelimiters, bool openedFile)
{
    Scanner scan;

    assert(delimiters != NULL);
    if (infile == NULL) return NULL;
    scan = (Scanner) GetBlock(sizeof *scan);
    scan->file = infile;
    scan->openedFile = openedFile;
    scan->delimiters = strdup(delimiters);
    scan->discardDelimiters = discardDelimiters;
    return scan;
}

/*
* Function: FreeScanner
* -----------------------
* This function closes the opened file, if NewScannerFromFilename was
* called. It frees the pointer to s->delimiters, and the scanner itself.
*/

void FreeScanner(Scanner s)
{
    if (s->openedFile) fclose(s->file);
    free (s->delimiters);
    free (s);
}

/*
* Function: ReadNextToken
* ------------------------
* This function gets the next character in the file, and creates
* a single-character string from that character. If the character
* is a member of the delimiter set, it checks if the function
* should discardDelimiters. If it shouldn't, it returns the boolean
* value true, and the single-character string in the buffer.
* If the function discards delimiters, it keeps on getting the next
* character from the file until it encounters EOF or a non-delimiter
* character. When the function finds a non-delimter, it passes the
* three arguments to ReadNonDelimiters function. If the buffer
* only consists of EOF (or delimiters if delimiters are discarded),
* it returns false. If the buffer has a valid word, it returns true.
*/

bool ReadNextToken(Scanner s, char buffer[], int bufLen)
{
    int next;
    char ch;

    assert(bufLen > 1);
    while (true)
    {
        next = fgetc(s->file);
        if (next == EOF ) return false;
        ch = next;
        buffer[0] = ch;
        buffer[1] = '\0';
        if (!IsInSet(ch, s->delimiters)) break;
        if (!s->discardDelimiters) return true;
    }
    return(ReadNonDelimiters(s, buffer, bufLen));
}

/*
* Function: ReadNonDelimiters
* ----------------------------
* This function reads non-delimiters and constructs a word until
* it gets EOF or non-delimiter in the file. When it does, or
* the length of the word is greater than bufLen, it puts the
* newest character back into the file (i.e. delimiter, EOF, or
* the (bufLen)th character in the buffer), and returns true.
*/

static bool ReadNonDelimiters(Scanner s, char buffer[], int bufLen)
{
    int i, next;
    char ch;

    i = 1;
    while (i < bufLen - 1)
    {
        next = fgetc(s->file);
        if (next == EOF) return PutCharBack(s, buffer, next, i);
        ch = next;
        if (IsInSet(ch, s->delimiters)) return PutCharBack(s, buffer, next, i);
        buffer[i++] = ch;
    }
    buffer[i] = '\0';
    return true;
}

/*
* Function: PutCharBack
* -----------------------
* This function pushes "next" back into the buffer string. It
* places '\0 at the end of the buffer to make the set of
* characters in the buffer a string. Then, it returns true,
* which shows that the characters in the buffer is a valid token.
*/

static bool PutCharBack(Scanner s, char buffer[], int next, int i)
{
    ungetc(next, s->file);
    buffer[i] = '\0';
    return true;
}

/*
* Function: SkipOver
* -------------------
* This function keeps on getting characters from the file
* as long as the characters are the ones in the skipSet.
* If it gets a character not in the set, it pushes that
* character back to the file, and returns the integer value
* of that character. This is a SkipOver set, sot he bool
* value passed in is true.
*/

int SkipOver(Scanner s, const char *skipSet)
{
    return( Skip(s, skipSet, true));
}

/*
* Function: SkipUntil
* -------------------
* If it gets a character not in the set, it pushes that
* character back to the file, and returns the integer value
* of that character. This is not a SkipOver set, so the bool
* value passed in is false.
*/

int SkipUntil(Scanner s, const char *untilSet)
{
    return(Skip(s, untilSet, false));
}

/*
* Function: Skip
* ---------------
* This function is a wrapping function for SkipOver and SkipUntil
* functions. See the descriptions for those functions for details.
* The third argument passed into this function, over, is set to
* true, if it is a wrapper for SkipOver, and set to false, if
* it is a wrapper for SkipUntil function.
*/

static int Skip(Scanner s, const char *set, bool over)
{
    int next;
    char ch;

    while (true)
    {
        next = fgetc(s->file);
        if (next == EOF) return EOF;
        ch = next;
        if ((over && !IsInSet(ch, set))||(!over && IsInSet(ch, set))) break;
    }
    ungetc(ch, s->file);
    return next;
}

/*
* Function: IsInSet
* ------------------
* This function checks if the given ch is in the given character
* set. It returns a boolean value accordingly.
*/

static bool IsInSet(char ch, const char *set)
{
    int i, len;

    assert (set != NULL);
    len = strlen (set);
    for (i = 0; i < len; i++)
    {
        if (ch == set[i]) return true;
    }
    return false;
}

/*
* Function: GetBlock
* Usage: result = GetBlock (sizeof(...));
* ---------------------------------------
* This function is similar to the GetBlock function in
* The Art and Science of C, p.675, by Eric S. Roberts.
*/

static void* GetBlock(int nbytes)
{
    void *result;

    result = malloc(nbytes);
    assert(result != NULL);
    return result;
}
